import java.util.Vector; 

/* Copyright K. Lano 2015-2019
   Package: QVT
*/ 

public class ObjectTemplateExp extends TemplateExp
{ Entity referredClass;
  Vector part = new Vector(); // PropertyTemplateItem

  public ObjectTemplateExp(Attribute att, Expression ex, Entity e)
  { super(att,ex); 
    referredClass = e; 
  }

  public ObjectTemplateExp(Attribute att, Entity e)
  { super(att,null); 
    referredClass = e; 
  }

  public Object clone()
  { Expression wclone = null; 
    if (where != null) 
    { wclone = (Expression) where.clone(); } 
    ObjectTemplateExp res = new ObjectTemplateExp(bindsTo,wclone,referredClass);
    for (int i = 0; i < part.size(); i++) 
    { PropertyTemplateItem pti = (PropertyTemplateItem) part.get(i); 
      res.addpart((PropertyTemplateItem) pti.clone()); 
    } 
    return res; 
  } 

  public void addpart(PropertyTemplateItem p)
  { part.add(p); }

  public void setpart(Vector parts)
  { part = parts; 
    if (parts == null) 
    { part = new Vector(); }        
  } 

  public void addPTI(Attribute att, Object v)
  { for (int i = 0; i < part.size(); i++) 
    { PropertyTemplateItem pti = (PropertyTemplateItem) part.get(i); 
      if (pti.referredProperty.getName().equals(att.getName()))
      { return; } 
    }       
    PropertyTemplateItem newpti = new PropertyTemplateItem(att);
    if (v instanceof ObjectTemplateExp)
    { newpti.setTemplate((ObjectTemplateExp) v); } 
    else if (v instanceof Expression) 
    { newpti.setValue((Expression) v); }  
    part.add(newpti); 
  } 

  public Object getPTI(Attribute att) 
  { for (int i = 0; i < part.size(); i++) 
    { PropertyTemplateItem pti = (PropertyTemplateItem) part.get(i); 
      if (pti.referredProperty.getName().equals(att.getName()))
      { if (pti.template != null) 
        { return pti.template; }
        else 
        { return pti.value; } 
      }  
    }   
    return null; 
  }     
      
  public Entity getEntity()
  { return referredClass; }

  public void setConstraint(Expression w)
  { where = w; } 

  public void addWhere(Expression w)
  { if (where == null) 
    { where = w; }
    else 
    { where = Expression.simplify("&", where, w, new Vector()); } 
  }  

  public Vector getObjectTemplateAttributes(Vector vars)
  { Vector res = new Vector(); 
    
    String dname = bindsTo.getName(); 
    for (int i = 0; i < vars.size(); i++) 
    { String v = vars.get(i) + ""; 
      if (dname.equals(v))
      { res.add(bindsTo); } 
    } 
    

    for (int i = 0; i < part.size(); i++) 
    { PropertyTemplateItem p = (PropertyTemplateItem) part.get(i);
      res.addAll(p.getObjectTemplateAttributes(vars)); 
    } 
    return res; 
  } 

  public Vector objectTemplateAttributes()
  { Vector res = new Vector(); 
    
    res.add(bindsTo);  
    
    for (int i = 0; i < part.size(); i++) 
    { PropertyTemplateItem p = (PropertyTemplateItem) part.get(i);
      res.addAll(p.objectTemplateAttributes()); 
    } 
    return res; 
  } 

  public boolean typeCheck(Vector types, Vector entities, Vector contexts, Vector env)
  { Vector context = new Vector(); 
    if (referredClass != null) 
    { context.add(referredClass); } 
    context.addAll(contexts); 
    // env.add(bindsTo)? 

    for (int i = 0; i < part.size(); i++) 
    { PropertyTemplateItem p = (PropertyTemplateItem) part.get(i);
      p.typeCheck(types,entities,context,env); 
    } 

    if (where != null) 
    { where.typeCheck(types,entities,context,env); } 

    return true; 
  } 

  public Expression toGuardCondition(Vector bound, Expression contextObj, Expression post)
  { Expression res = new BasicExpression(true);
 
    String bind = bindsTo.getName(); 
        
    boolean alreadybound = bound.contains(bind); 
    // System.out.println("BOUND = " + bound + " " + bindsTo); 

    Vector localbound = new Vector(); 
    localbound.addAll(bound); 
    
    if (alreadybound) { } 
    else 
    { localbound.add(bind); } 

    res = Expression.simplify("&",where,post,null);

    for (int i = 0; i < part.size(); i++)
    { PropertyTemplateItem p = (PropertyTemplateItem) part.get(i);
      p.setEntity(referredClass);
      res = p.toGuardCondition(localbound, contextObj, res);
    }

    Expression btexp = new BasicExpression(bindsTo);
    // Expression scope = new BasicExpression(referredClass);
    // Expression decl = new BinaryExpression(":", btexp, scope); 
    // res = Expression.simplify("&",decl,res,null);

    if (alreadybound)    
    { return res; } 

    Expression scope = new BasicExpression(referredClass);
    Expression decl = new BinaryExpression(":", btexp, scope); 
    res = new BinaryExpression("#", decl, res);

    return res;
  }

  public Expression toSourceExpression(Vector bound, Expression contextObj)
  { Expression res = new BasicExpression(true);
 
    String bind = bindsTo.getName(); 
    if (bound.contains(bind)) { } 
    else 
    { bound.add(bindsTo.getName()); }  // Surely? 

    for (int i = 0; i < part.size(); i++)
    { PropertyTemplateItem p = (PropertyTemplateItem) part.get(i);
      p.setEntity(referredClass);
      Expression pexp = p.toSourceExpression(bound, contextObj);
      res = Expression.simplify("&",res,pexp,null);
    }

    res = Expression.simplify("&",res,where,null);
    // Expression btexp = new BasicExpression(bindsTo);
    // Expression scope = new BasicExpression(referredClass);
    // Expression decl = new BinaryExpression(":", btexp, scope); 
    // res = Expression.simplify("&",decl,res,null);
    return res;
  }

  public Expression toUMLRSDSantecedent()
  { Vector bnd = new Vector(); 
    return toUMLRSDSantecedent(bnd, null); 
  }     

  public Expression toUMLRSDSantecedent(Vector bound, Expression contextObj)
  { Expression res = new BasicExpression(true);
 
    String bind = bindsTo.getName(); 
    if (bound.contains(bind)) { } 
    else 
    { bound.add(bindsTo.getName()); }  // Surely? 

    for (int i = 0; i < part.size(); i++)
    { PropertyTemplateItem p = (PropertyTemplateItem) part.get(i);
      p.setEntity(referredClass);
      if (p.isManyValued())
      { Expression pexp = p.toSourceExpression0(bound, contextObj);
        res = Expression.simplify("&",res,pexp,null);
      } 
    }

    // res = Expression.simplify("&",res,where,null);
    // Expression btexp = new BasicExpression(bindsTo);
    // Expression scope = new BasicExpression(referredClass);
    // Expression decl = new BinaryExpression(":", btexp, scope); 
    // res = Expression.simplify("&",decl,res,null);
    return res;
  }

  public boolean isEmpty()
  { return part.size() == 0; }
  
  public Expression toTargetExpression(Vector bound, Expression contextObj, Expression setting)
  { Expression res = setting; 
    Expression btexp = new BasicExpression(bindsTo);
    boolean alreadybound = bound.contains(bindsTo + ""); 
    // System.out.println("BOUND = " + bound + " " + bindsTo); 

    if (alreadybound) { } 
    else 
    { bound.add(bindsTo + ""); } 

    for (int i = 0; i < part.size(); i++)
    { PropertyTemplateItem p = (PropertyTemplateItem) part.get(i);
      p.setEntity(referredClass);
      Expression pexp = p.toTargetExpression(bound, btexp,null);
      res = Expression.simplify("#&",res,pexp,null);
    }

    res = Expression.simplify("#&",res,where,null);

    if (alreadybound)    
    { return res; } 

    Expression scope = new BasicExpression(referredClass);
    Expression decl = new BinaryExpression(":", btexp, scope); 
    res = new BinaryExpression("#", decl, res);

    // System.out.println("TARGET ACTION for " + this + " IS " + res); 

    return res;
  }

  public Expression toTargetExpressionOp(Vector bound, Expression contextObj, Expression setting)
  { Expression res = setting; 
    Expression btexp = new BasicExpression(bindsTo);
    boolean alreadybound = bound.contains(bindsTo + ""); 
    // System.out.println("BOUND = " + bound + " " + bindsTo); 

    if (alreadybound) { } 
    else 
    { bound.add(bindsTo + ""); } 

    for (int i = 0; i < part.size(); i++)
    { PropertyTemplateItem p = (PropertyTemplateItem) part.get(i);
      p.setEntity(referredClass);
      Expression pexp = p.toTargetExpressionOp(bound, btexp, null);
      res = Expression.simplify("#&",res,pexp,null);
    }
    res = Expression.simplify("#&",res,where,null);
    if (alreadybound)
    { return res; } 
    Expression scope = new BasicExpression(referredClass);
    Expression decl = new BinaryExpression(":", btexp, scope); 
    res = new BinaryExpression("#", decl, res);

    System.out.println("TARGET ACTION for " + this + " IS " + res); 
    return res;
  }


  public String toString()
  { String res = bindsTo + " : " + referredClass + " { ";
    for (int i = 0; i < part.size(); i++)
    { PropertyTemplateItem p = (PropertyTemplateItem) part.get(i);
      res = res + p;
      if (i < part.size() - 1)
      { res = res + ", "; }
    }
    res = res + " }"; 
    if (where != null) 
    { res = res + " { " + where + " }"; } 
    return res;
  } // and { where }

  public String toQVTO(Attribute src, Map whens)
  { Vector bound = new Vector(); 
    // ignore the first target - this will be result variable 

    String res = ""; 
    Expression contxt = new BasicExpression("result"); 
    contxt.setType(new Type(referredClass)); 
    contxt.setElementType(new Type(referredClass)); 

    // Expression scope = new BasicExpression(referredClass);

    for (int i = 0; i < part.size(); i++)
    { PropertyTemplateItem p = (PropertyTemplateItem) part.get(i);
      res = res + p.toQVTO(contxt,src,bound,whens);
      // if (i < part.size() - 1)
      // { res = res + " "; }
    }
    return res; 
  }     

  public String toQVTO(Expression contextObject, Attribute src, Vector bound, Map whens)
  { String res = " object " + referredClass + " { ";

    // Expression scope = new BasicExpression(referredClass);

    for (int i = 0; i < part.size(); i++)
    { PropertyTemplateItem p = (PropertyTemplateItem) part.get(i);
      res = res + p.toQVTO(contextObject,src,bound,whens);
      // if (i < part.size() - 1)
      // { res = res + "; "; }
    }
    res = res + " }"; 
    // if (where != null) 
    // { res = res + " { " + where + " }"; } 
    return res;
  } // and { where }

  public String toQVTOnoobject(Expression contextObject, Attribute src, Vector bound, Map whens)
  { String res = "  "; 
    Expression scope = new BasicExpression(referredClass);

    for (int i = 0; i < part.size(); i++)
    { PropertyTemplateItem p = (PropertyTemplateItem) part.get(i);
      res = res + p.toQVTO(contextObject,src,bound,whens);
      // if (i < part.size() - 1)
      // { res = res + "; "; }
    }
    return res;
  } // and { where }

  public Expression toUMLRSDS(Map whens)
  { Expression res = new BasicExpression(true); 
    
    Vector bound = new Vector(); 
    Expression btexp = new BasicExpression(bindsTo); 
    bound.add(bindsTo + "");

    for (int i = 0; i < part.size(); i++)
    { PropertyTemplateItem p = (PropertyTemplateItem) part.get(i);
      p.setEntity(referredClass);
      Expression pexp = p.toUMLRSDS(bound, btexp, null, whens);
      res = Expression.simplify("#&",res,pexp,null);
    }

    return res; 
  } 

  public Expression toUMLRSDS(Vector bound, Expression contextObj, Expression setting, Map whens)
  { Expression res = setting; 
    Expression btexp = new BasicExpression(bindsTo);
    boolean alreadybound = bound.contains(bindsTo + ""); 
    // System.out.println("BOUND = " + bound + " " + bindsTo); 

    if (alreadybound) { } 
    else 
    { bound.add(bindsTo + ""); } 

    for (int i = 0; i < part.size(); i++)
    { PropertyTemplateItem p = (PropertyTemplateItem) part.get(i);
      p.setEntity(referredClass);
      Expression pexp = p.toUMLRSDS(bound, btexp, null, whens);
      res = Expression.simplify("#&",res,pexp,null);
    }

    res = Expression.simplify("#&",res,where,null);

    if (alreadybound)    
    { return res; } 

    Expression scope = new BasicExpression(referredClass);
    Expression decl = new BinaryExpression(":", btexp, scope); 
    res = new BinaryExpression("#", decl, res);

    // System.out.println("TARGET ACTION for " + this + " IS " + res); 

    return res;
  }

  public Expression toUMLRSDSroot(String contextObj)
  { // Expression res = setting; 
    BasicExpression btexp = new BasicExpression(bindsTo);

    BasicExpression scope = new BasicExpression(referredClass);
   
    BasicExpression cid = new BasicExpression(contextObj.toLowerCase() + "Id"); 
    scope.setArrayIndex(cid); 
    Expression decl = new BinaryExpression("=", btexp, scope); 
    return decl;
  }


  public int complexity() 
  { int res = 3; 
    for (int i = 0; i < part.size(); i++)
    { PropertyTemplateItem p = (PropertyTemplateItem) part.get(i);
      res = res + p.complexity();
    }
    if (where != null) 
    { res = res + where.syntacticComplexity(); } 
    return res;
  } 

  public int epl()
  { int res = 1; 
    for (int i = 0; i < part.size(); i++)
    { PropertyTemplateItem p = (PropertyTemplateItem) part.get(i);
      res = res + p.epl();
    }
    return res; 
  }     

  public Vector operationsUsedIn() 
  { Vector res = new Vector(); 
    for (int i = 0; i < part.size(); i++)
    { PropertyTemplateItem p = (PropertyTemplateItem) part.get(i);
      res.addAll(p.operationsUsedIn());
    }
    if (where != null) 
    { res.addAll(where.allOperationsUsedIn()); } 
    return res;
  } 

}


